/*  g_sysv_shared.c - SystemV IPC SHARED MEMORY TEST
 *  
 *  (C) Copyright XII 2004 Michal 'GiM' Spadlinski <gim at skrzynka dot pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "g_sysv_shared.h"
#include "dynstuff.h"
#include "g_ipc_wspolne.h"

extern const int g_data_format_s;
extern struct g_data_format parentp[2], childp[2];
extern char *memp;
int showtimes(char *str, list_t p);

int do_pause = 1;
void dummy_ignore(int s)
{
	signal(SIGUSR1, SIG_IGN);
	do_pause = 0;
	return;
}

int prefork_sysv_shared(void)
{
	signal(SIGUSR1, dummy_ignore);
	return 0;
}

int child_sysv_shared(int off)
{
	int shared, i;
	long *lp = NULL;
	char *p, *q;
	struct timeval s, k;
	list_t sending = NULL, receiving = NULL;

	DEB("[%d] in child\n", getpid());
	for (i=0; i<MAX_OPEN_TRIES; i++)
	{
		if ((shared = shmget(G_KEY, 20 KB, 0666)) != -1)
			break;
		SLEEP(s, 500);
	}

	if (i == MAX_OPEN_TRIES)
	{
		perror ("child: creating shared");
		return -1;
	}
	DEB("[%d] child shared opened\n", getpid());

	if ((p = shmat(shared, 0, 0)) == (void *)-1)
	{
		perror("child attaching shared");
		return -1;
	}
	DEB("[%d] child shared attached %08X\n", getpid(), (unsigned int)p);

	if(do_pause)
	{
		pause();
		DEB("[%d] child signaled\n", getpid());
	}

	childp[off+0].type = 1;
	childp[off+0].id = getpid();
	q = childp[off+0].p;
	for (i=0; i<IPCQ_S; i++)
		*q++ = 'A';

	q = p + 10 KB;
	lp = (long *)p;

	DEB("[%d] child shared data prepared\n", getpid());

	for (i=0; i<MAX_QUERIES; i++)
	{
		TIMESTAMP(s);
		DEB("[%d] child sending data\n", getpid());
		memcpy(q + long_s, &(childp[off+0].id), IPCQ_S + long_s);
		memcpy(q, &(childp[off+0].type), long_s);
		DEB("[%d] child data sent %08lX\n", getpid(), *(unsigned long*)q);
		TIMESTAMP(k);

		DIFFTIME(k, s);
		list_add(&sending, &k, sizeof(struct timeval));

		TIMESTAMP(s);
		while (*lp != getpid()) sched_yield();
		memcpy (&(childp[off+1]), lp, IPCQ_S + 2 * long_s);
		*lp = 0;

		TIMESTAMP(k);
		DIFFTIME(k, s);
		list_add(&receiving, &k, sizeof(struct timeval));

		childp[off+1].p[IPCQ_S - 1] = 0; 
		DEB("[%d] child received data from %li\n", getpid(), childp[off+1].id);
	}
	showtimes("[+] sysv sh memory   sending", sending);
	showtimes("[+] sysv sh memory receiving", receiving);

	shmdt(p);
	DEB("[%d] child detached\n", getpid());

	list_destroy(sending, 1);
	list_destroy(receiving, 1);

	return 0;
}

int parent_sysv_shared(int pid)
{
	int shared;
	struct shmid_ds shmid;
	char *p;
	long *lp;

	if ((shared = shmget(G_KEY, 20 KB, IPC_CREAT|0666|IPC_EXCL)) == -1)
	{
		perror ("parent creating shared");
		return -1;
	}
	DEB("[%d] parent shared created\n", getpid());

	if ((p = shmat(shared, 0, 0)) == (void *)-1)
	{
		perror("parent attaching shared");
		return -1;
	}
	memp = p;
	lp = (long *)p;
	DEB("[%d] parent shared attached %08X\n", getpid(), (unsigned int)p);
	while ((char *)(lp) < memp + 20 KB)
		*lp++ = 0;
	DEB("[%d] parent memory cleared\n", getpid());
	kill(pid, SIGUSR1);
	DEB("[%d] parent sognal send\n", getpid());

	if (shmctl(shared, IPC_STAT, &shmid) == -1)
	{
		perror("parent stat shared");
		return -1;
	}
	DEB("[%d] parent shared stated\n", getpid());

	lp = (long *)(memp + 10 KB);
	while (1)
	{
		DEB("[%d] parent shared waiting for data %ld\n", 
				getpid(), *lp);
		if (*lp == 1)
		{
			/* receive */
			memcpy (&(parentp[0]), lp, IPCQ_S + 2 * long_s);
			*lp = 0;
			DEB("[%d] parend shared data received from %li",
					getpid(), parentp[0].id);

			/* send */
			parentp[0].type = parentp[0].id;
			parentp[0].id = getpid();
			memcpy (memp + long_s, &(parentp[0].id), IPCQ_S + long_s);
			memcpy (memp, &(parentp[0].type), long_s);
		} else {
			if (waitpid(pid, NULL, WNOHANG) != 0)
				break;
			sched_yield();
		}
	}

	DEB("[%d] parent child exited\n", getpid());
	shmdt(memp);
	DEB("[%d] parent detached\n", getpid());
	DEB("%zu %lu\n", shmid.shm_segsz, shmid.shm_nattch);

	if (shmctl(shared, IPC_RMID, NULL) == -1)
	{
		perror("parent removing shared");
		return -1;
	}

	return 0;
}
